/*************************************************************************\
|* inpcon.c -- Input connection source file for Network Toolkit Library. *|
|* (c)1997 Project INSANITY Software -- By Steve Conley.                 *|
\*************************************************************************/

/* System includes */
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <fcntl.h>
#include <unistd.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <signal.h>

/* Project includes */
#include "nettool.h"

/* Project functions */
int nonblock(int s); /* Sets socket s nonblocking */
int descrwrite(int desc,char* text); /* Writes text to a descriptor */


/* Set a socket nonblocking -- a simple fcntl call */
int nonblock(int s){
#ifdef BLOCKING
	int flags;
	flags=fcntl(s,F_GETFL);
	flags |= O_NONBLOCK;
	if(fcntl(s,F_SETFL,flags) < 0){
#ifndef QUIET_LIBRARY
		printf("COMMLIB: comm.c: FATAL ERROR!  Nonblock failed!\n");
#endif
		return 0;
	}
	return 1;
#endif
	return (int) 1;
}

/* Initialize an input connection */
INPCON* inpcon_init(int port){
	INPCON* newinpcon;			/* The input connection structure */
	char    host[MAXHOSTNAME+1];	/* Hostname temp variable         */
	struct  sockaddr_in addy;	/* Address information temp space */
	char*   opt;				/* Temp character varibable       */
	struct  hostent *hp;		/* Temporary host information     */
	int     sock;				/* Socket temp location           */
/* Now, initialize the address structure */
	bzero(&addy,sizeof(struct sockaddr_in));
/* Get the host address of the current machine */
	gethostname(host,MAXHOSTNAME);
	hp=gethostbyname(host);
	if(hp==NULL){
#ifndef QUIET_LIBRARY
		printf("NetTool: inpcon_init: Machine has no hostname? error.\n");
#endif
		return (INPCON*) NULL;
	}
/* Fill in needed data entries */
	addy.sin_family=hp->h_addrtype;
	addy.sin_port=htons(port);
/* Attempt to open a socket */
	sock=socket(AF_INET,SOCK_STREAM,0);
	if(sock < 0){
#ifndef QUIET_LIBRARY
		printf("NetTool: inpcon_init: Socket open failed.\n");
#endif
		return (INPCON*) NULL;
	}
/* Set it up for local reuse */
	if(setsockopt(sock,SOL_SOCKET,SO_REUSEADDR,(char *)&opt,sizeof(opt)) < 0){
#ifndef QUIET_LIBRARY
		printf("NetTool: inpcon_init: Set socket option failed--server error.\n");
#endif
		return (INPCON*) NULL;
	}
/* Bind socket to port */
	if(bind(sock,(struct sockaddr *) & addy,sizeof(addy)) < 0){
#ifndef QUIET_LIBRARY
		printf("NetTool: inpcon_init: Could not bind socket to port--port in use.\n");
#endif
		return (INPCON*) NULL;
	}
	nonblock(sock);
	listen(sock,QUEUE_LEN);
	newinpcon=(INPCON*)malloc(sizeof(INPCON));
	if(newinpcon==NULL){
#ifndef QUIET_LIBRARY
		printf("NetTool: inpcon_init: Out of memory error.\n");
#endif
		return (INPCON*) NULL;
	}
	newinpcon->port=port;
	newinpcon->master_socket=sock;
	newinpcon->numconnects=0;
	return (INPCON*) newinpcon;
}

/* Accept a new connection */
int inpcon_accept(INPCON* connections){
/* Variables that we need */
	struct sockaddr_in addr;
	struct hostent*    host;
	CON*   cur;
	CON*   prv;
	int    i,j;
	i=sizeof(addr); /* Get the size of the address structure */
	getsockname(connections->master_socket,(struct sockaddr*)&addr,&i);
	if((i=accept(connections->master_socket,(struct sockaddr*)&addr,&i)) < 0){
		return(0);
	}
	nonblock(i);
	cur=connections->head;
	prv=NULL;
	while(cur!=NULL){
		prv=cur;
		cur=cur->next;
	}
	connections->numconnects++;
	if(prv==NULL){
		connections->head=(CON*)malloc(sizeof(CON));
		connections->head->descr=i;
		cur=connections->head;
	}else{
		prv->next=(CON*)malloc(sizeof(CON));
		prv->next->descr=i;
		cur=prv->next;
	}
	cur->next=NULL;
	j=sizeof(addr);
	if(getpeername(cur->descr,(struct sockaddr*)&addr,&j) < 0){
#ifndef QUIET_LIBRARY
		printf("Nettool: inpcon_accept: Couldn't get connection peer's host info\n");
#endif
		return(1);
	}
	host=gethostbyaddr((char *)&addr.sin_addr,sizeof(addr.sin_addr), AF_INET);
	if(host==NULL){
#ifndef QUIET_LIBRARY
		printf("Nettol: inpcon_accept: Hostname lookup failed.\n");
#endif
		return(1);
	}
	strncpy(cur->host,host->h_name,MAXHOSTNAME-1);
	cur->host[MAXHOSTNAME]='\0';
	return(1);
}

/* Send a message over a descriptor. */
int consend(CON* connection,char* message){
	return descrwrite(connection->descr,message);
}

char* conrecv(CON* connection){
	char  buffer[8192]; /* 8K buffer */
	char* buf;
	int   length;
	bzero(buffer,8192);
	length=read(connection->descr,buffer,8192);
	if(length==-1){
		return NULL;
	}
	buffer[8191]='\0';
	buf=(char*)malloc(strlen(buffer)+1);
	strcpy(buf,buffer);
	return(buf);
}

/* Disconnect a connection */
int disconnect(INPCON* list,CON* connection){
	CON* cur;
	CON* prv;
	prv=NULL;
	cur=list->head;
	while(cur!=connection){
		prv=cur;
		cur=cur->next;
	}
	if(cur==NULL){
		return(0);
	}
	list->numconnects--;
	if(prv==NULL){
		cur=list->head;
		list->head=list->head->next;
		close(cur->descr);
		cur->next=NULL;
		free(cur);
		return(1);
	}
	prv->next=cur->next;
	close(cur->descr);
	free(cur);
	return(1);
}

/* Destroy an INPCON table */
void INPdestroy(INPCON* table){
	CON* ptr;
	CON* prv;
	ptr=table->head;
	while(ptr!=NULL){
		prv=ptr;
		ptr=ptr->next;
		close(prv->descr);
		free(prv);
	}
	close(table->master_socket);
	free(table);
	table=NULL;
}

int descrwrite(int desc,char* text){
	int thisfar,thisround,total;
	total=strlen(text);
	thisfar=0;
	signal(SIGPIPE,SIG_IGN);
	do{
		thisround=write(desc,text+thisfar,total-thisfar);
		if(thisround < 0){
#ifndef QUIET_LIBRARY
			printf("Nettool: descrwrite: Write failed.\n");
#endif
			return(0);
		}
		thisfar+=thisround;
	}while(thisfar < total);
	return(1);
}
